local _G = getfenv(0)
-- Local versions for performance
local string = _G.string
local math = _G.math
local pairs = _G.pairs
local LibStub = _G.LibStub
local floor, ceil, min, max, abs = math.floor, math.ceil, math.min, math.max, math.abs
local tconcat = table.concat
local wipe = _G.wipe

local CompactRunes = LibStub("AceAddon-3.0"):NewAddon("Compact Runes", "AceConsole-3.0", "AceEvent-3.0","AceTimer-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("CompactRunes", true)
local LDB = LibStub("LibDataBroker-1.1")
local LibQTip = LibStub("LibQTip-1.0")
local icon = LibStub("LibDBIcon-1.0")
local LSM = LibStub:GetLibrary("LibSharedMedia-3.0")

-- Try to remove the Git hash at the end, otherwise return the passed in value.
local function cleanupVersion(version)
	local iter = string.gmatch(version, "(.*)-[a-z0-9]+$")
	if iter then
		local ver = iter()
		if ver and #ver >= 3 then
			return ver
		end
	end
	return version
end

local ADDON_NAME = ...
local ADDON_VERSION = cleanupVersion("0.2")

-- Define a simplistic class for combat statistics
CombatStats = {}
CombatStats.__index = CombatStats

function CombatStats:new()
    local stats = {}
    setmetatable(stats, CombatStats)
    stats:Reset()
    return stats
end

function CombatStats:Reset()
	self.rcProcs = 0
	self.reProcs = self.reProcs or {}
	wipe(self.reProcs)
	self.regenUses = self.regenUses or {}
	wipe(self.regenUses)
end

local RuneColors = {
	["BLOOD"] = { r = 0.77, g = 0.12, b = 0.23, a = 1.0 },
	["UNHOLY"] = { r = 0.3, g = 0.8, b = 0.1, a = 1.0 },
	["FROST"] = { r = 0.0, g = 0.4, b = 0.7, a = 1.0 },
	["DEATH"] = { r = 0.51, g = 0.23, b = 0.65, a = 1.0 },
}

local RuneCooldownColors = {
	["BLOOD"] = { r = 1.0, g = 0.0, b = 0.0, a = 0.1 },
	["UNHOLY"] = { r = 0.0, g = 1.0, b = 0.0, a = 0.1 },
	["FROST"] = { r = 0.0, g = 0.0, b = 1.0, a = 0.1 },
	["DEATH"] = { r = 1.0, g = 0.0, b = 1.0, a = 0.1 },
}

local RuneMapping = {
	[1] = "BLOOD",
	[2] = "UNHOLY",
	[3] = "FROST",
	[4] = "DEATH",
}

local HEXFMT = "|c%2x%2x%2x%2x"
local function chatColorCode(color)
	local a = 255
	local r = ceil(255 * color.r)
	local g = ceil(255 * color.g)
	local b = ceil(255 * color.b)
	return HEXFMT:format(a, r, g, b)
end

local RuneChatColors = {}
for k, v in pairs(RuneColors) do
	RuneChatColors[k] = chatColorCode(v)
end

-- Define Bar for now but the rest is at the bottom of the file.
local Bar = {}

CompactRunes.bars = {}
CompactRunes.playerName = UnitName("player")
CompactRunes.isDK = false
CompactRunes.TotalStats = CombatStats:new()
CompactRunes.LastFightStats = CombatStats:new()
CompactRunes.regenMethod = "None"

local Broker = CreateFrame("Frame")
Broker.obj = LDB:NewDataObject(GetAddOnMetadata(ADDON_NAME, "Title"), {
    type = "data source",
    icon = "Interface\\Icons\\Spell_DeathKnight_DeathStrike",
    label = GetAddOnMetadata(ADDON_NAME, "Title"),
    text = GetAddOnMetadata(ADDON_NAME, "Title"),
    barValue = 0,
    barR = 0,
    barG = 0,
    barB = 1,
	OnClick = function(clickedframe, button)
		if button == "RightButton" then
			local optionsFrame = InterfaceOptionsFrame

			if optionsFrame:IsVisible() then
				optionsFrame:Hide()
			else
				CompactRunes:ShowOptions()
			end
		elseif button == "LeftButton" and IsShiftKeyDown() then
		    CompactRunes.TotalStats:Reset()
		    CompactRunes.LastFightStats:Reset()
        end
	end
} )

local GREEN = "|cff00ff00"
local YELLOW = "|cffffff00"
local BLUE = "|cff0198e1"
local ORANGE = "|cffff9933"
local addonHdr = GREEN.."%s %s"
local totalDataHdr = ORANGE..L["Session"]
local lastFightValueHdr = ORANGE..L["Last Fight"]
local reProcFmt = "%s%d|r / %s%d|r / %s%d|r / %s%d"
local valuePlusPercFmt = "%d (%.1f%%)"
local reProcTypeHdr = "    %s%s"

local function valuePlusPercent(value, total)
	local val = value or 0
	local tot = total or 0
	local perc = (tot > 0) and (val / tot * 100) or 0
	return valuePlusPercFmt:format(val, perc)
end

local function AddStats(tooltip, stats)
	local totalProcs = 0
	local reProcs = stats.reProcs
	for k, v in pairs(reProcs) do
		totalProcs = totalProcs + (v or 0)
	end
	local reRegenUses = stats.regenUses["RE"] or 0
	local rcRegenUses = stats.regenUses["RC"] or 0

    tooltip:AddSeparator(1)
	tooltip:AddLine(L["Runic Corruption Procs"],
		valuePlusPercent(stats.rcProcs, rcRegenUses))
	tooltip:AddLine(L["Runic Empowerment Procs"],
		valuePlusPercent(totalProcs, reRegenUses))
	tooltip:AddLine(
		reProcTypeHdr:format(RuneChatColors["BLOOD"],"Blood"),
			valuePlusPercent(reProcs[1], totalProcs))
	tooltip:AddLine(
		reProcTypeHdr:format(RuneChatColors["UNHOLY"],"Unholy"),
			valuePlusPercent(reProcs[2], totalProcs))
	tooltip:AddLine(
		reProcTypeHdr:format(RuneChatColors["FROST"],"Frost"),
			valuePlusPercent(reProcs[3], totalProcs))
	tooltip:AddLine(
		reProcTypeHdr:format(RuneChatColors["DEATH"],"Death"),
			valuePlusPercent(reProcs[4], totalProcs))


	--tooltip:AddLine("RE Procs By Type",
	--	reProcFmt:format(RuneChatColors["BLOOD"], (reProcs[1] or 0), 
	--		RuneChatColors["UNHOLY"], (reProcs[2] or 0), 
	--		RuneChatColors["FROST"], (reProcs[3] or 0), 
	--		RuneChatColors["DEATH"], (reProcs[4] or 0)))
end

function Broker.obj:OnEnter()
	local tooltip = LibQTip:Acquire("CompactRunesTooltip", 2, "LEFT", "RIGHT")
	self.tooltip = tooltip 

    tooltip:AddHeader(addonHdr:format(GetAddOnMetadata(ADDON_NAME,"Title"), ADDON_VERSION))
    tooltip:AddLine()

    if CompactRunes.isDK then
        tooltip:AddLine(L["Shift + Left-Click to reset"], "", 1, 1, 1)
        tooltip:AddLine()

        tooltip:AddLine(totalDataHdr)
        AddStats(tooltip, CompactRunes.TotalStats)

        tooltip:AddLine()
	    tooltip:AddLine(lastFightValueHdr)
        AddStats(tooltip, CompactRunes.LastFightStats)	
    end

	tooltip:SmartAnchorTo(self)
	tooltip:Show()
end

function Broker.obj:OnLeave()
	LibQTip:Release(self.tooltip)
	self.tooltip = nil
end

local LookupOrKeyMT = {__index = function (t,k) return k end}
local SpellIds = {
	["Frost Fever"] = 59921,
	["Blood Plague"] = 59879,
	["Freezing Fog"] = 59052,
	["Killing Machine"] = 51128,
	["Sudden Doom"] = 49530,
	["Blood Charge"] = 114851,
	["Shadow Infusion"] = 91342,
	["Dark Transformation"] = 63560,
	["Death Coil"] = 47541,
	["Rune Strike"] = 56815,
	["Frost Strike"] = 49143,
	["Blood Tap"] = 45529,
	["Runic Empowerment"] = 81229,
	["Runic Corruption"] = 51460,
}
local SpellNames = {}
setmetatable(SpellNames, LookupOrKeyMT)
local function LoadSpellNames()
	for k,v in pairs(SpellIds) do
		if not SpellNames[k] then
			SpellNames[k] = (GetSpellInfo(SpellIds[v]))
		end
	end
end
LoadSpellNames()

local procsList = {
	"Freezing Fog", "Killing Machine", "Sudden Doom", "Blood Charge"
}

local frames = {}

local defaults = {
	profile = {
		minimap = {
			hide = true,
		},
		debug = false,
		locked = false,
		ooc_alpha = 0.4,
		rechargeAlpha = 0.3,
		runeRechargeNums = false,
		font_face = "Friz Quadrata TT",
		font_size = 12,
		fontFlags = {
			OUTLINE = true,
			THICKOUTLINE = false,
			MONOCHROME = false
		},
		texture = "CompactRunes",
		border = "Blizzard Tooltip",
		runeOrder = "BFU",
		width = 30,
		height = 10,
		spacing = 4,
		["frame"] = {
			x = 0,
			y = 0,
			color = { r = 0.1, g = 0.1, b = 0.1, a = 1.0 },
			bdcolor = { r = 0.5, g = 0.5, b = 0.5, a = 1.0 },
		},
		["runes"] = {
			inset = 6,
		},
		procs = true,
		enabled = {},
		-- Bars
		bars = {
			-- Provide defaults for all bars.
			-- These are inherited if no bar or no value is set.
			['**'] = {
				textcolor = { r = 1.0, g = 1.0, b = 1.0, a = 1.0 },
			},
			["RunicPowerBar"] = {
				enabled = true,
				color = { r = 0.0, g = 0.82, b = 1.0, a = 1.0 },
				bgcolor = { r = 0.0, g = 0.205, b = 0.25, a = 1.0 },
				textcolor = { r = 0.66, g = 0.82, b = 1.0, a = 1.0 },
			},
			["FrostFeverBar"] = {
				enabled = true,
				color = { r = 0.0, g = 0.45, b = 0.6, a = 1.0 },
				bgcolor = { r = 0.0, g = 0.1125, b = 0.15, a = 0.9 },
				textcolor = { r = 0.8, g = 0.8, b = 0.8, a = 1.0 },
			},
			["BloodPlagueBar"] = {
				enabled = true,
				color = { r = 0.5, g = 0.0, b = 0.6, a = 1.0 },
				bgcolor = { r = 0.125, g = 0.0, b = 0.15, a = 0.9 },
				textcolor = { r = 0.8, g = 0.8, b = 0.8, a = 1.0 },
			},
			["DarkTransformationBar"] = {
				enabled = true,
				color = { r = 0.0, g = 0.5, b = 0.0, a = 1.0 },
				bgcolor = { r = 0.125, g = 0.0, b = 0.15, a = 0.9 },
				textcolor = { r = 0.83, g = 0.66, b = 0.86, a = 1.0 },
				alt_color = { r = 0.5, g = 0.25, b = 0.0, a = 1.0 },
				alt_textcolor = { r = 0.8, g = 0.8, b = 0.8, a = 1.0 },
			},
		},
	}
}
for k, v in pairs(procsList) do
	defaults.profile.enabled[v] = true
end

function CompactRunes:GetOptions()
    if not options then
        options = {
            type = "group",
            name = GetAddOnMetadata(ADDON_NAME, "Title"),
            args = {
				core = self:GetGeneralOptions(),
				procs = self:GetProcsOptions()
            }
        }
		options.args.profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db)
    end
    return options
end

function CompactRunes:ShowOptions()
	InterfaceOptionsFrame_OpenToCategory(self.optionsFrame.Main)
end

function CompactRunes:GetGeneralOptions()
	local core = {
	    order = 1,
		name = L["General Options"],
		type = "group",
		args = {
		    description = {
		        order = 1,
		        type = "description",
		        name = L["CompactRunes_Desc"],
		    },
		    generalOptions = {
		        order = 100,
		        type = "header",
		        name = L["General Options"],
		    },
			locked = {
				name = L["Lock"],
				desc = L["LockDesc"],
				type = "toggle",
				order = 110,
				set = function(info, val)
				    self.db.profile.locked = val
					frames.main.lock = val
					frames.main:EnableMouse(not val)
				end,
                get = function(info)
					return self.db.profile.locked
				end,
			},
			runeOrder = {
				order = 120,
				name = L["Rune Order"],
				desc = L["RuneOrder_Desc"],
				type = "select",
				values = {
				    ["BFU"] = L["Blood, Frost, Unholy"],
				    ["UFB"] = L["Unholy, Frost, Blood"],
				    ["FBU"] = L["Frost, Blood, Unholy"],
				    ["BUF"] = L["Blood, Unholy, Frost"],
				},
				set = function(info, val)
				    self.db.profile.runeOrder = val
					self:SetRuneOrder(val)
				end,
                get = function(info)
                    return self.db.profile.runeOrder
                end,
			},
			runeRechargeNums = {
				name = L["Recharge Numbers"],
				desc = L["RechargeNumbersDesc"],
				type = "toggle",
				order = 130,
				set = function(info, val)
				    self.db.profile.runeRechargeNums = val
				end,
                get = function(info)
					return self.db.profile.runeRechargeNums
				end,
			},
			ooc_alpha = {
				order = 140,
				name = L["OOC Alpha"],
				desc = L["OOCAlpha_Desc"],
				type = "range",
				min = 0,
				max = 1,
				step = 0.05,
				set = function(info, val)
				    self.db.profile.ooc_alpha = val 
				end,
				get = function(info, val)
				    return self.db.profile.ooc_alpha
				end,					
			},
		    features = {
		        order = 200,
		        type = "header",
		        name = L["Features"],
		    },
			runicpower = {
				name = L["Runic Power"],
				desc = L["RunicPowerDesc"],
				type = "toggle",
				order = 210,
				set = function(info, val)
				    self.db.profile.bars.RunicPowerBar.enabled = val
				end,
                get = function(info)
					return self.db.profile.bars.RunicPowerBar.enabled
				end,
			},
			frostFever = {
				name = SpellNames["Frost Fever"],
				desc = SpellNames["Frost Fever"],
				type = "toggle",
				order = 220,
				set = function(info, val)
				    self.db.profile.bars.FrostFeverBar.enabled = val
				end,
                get = function(info)
					return self.db.profile.bars.FrostFeverBar.enabled
				end,
			},
			bloodPlague = {
				name = SpellNames["Blood Plague"],
				desc = SpellNames["Blood Plague"],
				type = "toggle",
				order = 230,
				set = function(info, val)
				    self.db.profile.bars.BloodPlagueBar.enabled = val
				end,
                get = function(info)
					return self.db.profile.bars.BloodPlagueBar.enabled
				end,
			},
			darktrans = {
				name = SpellNames["Dark Transformation"],
				desc = SpellNames["Dark Transformation"],
				type = "toggle",
				order = 240,
				set = function(info, val)
				    self.db.profile.bars.DarkTransformationBar.enabled = val
				end,
                get = function(info)
					return self.db.profile.bars.DarkTransformationBar.enabled
				end,
			},
			procs = {
				name = L["Procs"],
				desc = L["ProcsDesc"],
				type = "toggle",
				order = 250,
				set = function(info, val)
				    self.db.profile.procs = val
				end,
                get = function(info)
					return self.db.profile.procs
				end,
			},
		    dimensions = {
		        order = 300,
		        type = "header",
		        name = L["Dimensions"],
		    },
			width = {
				order = 310,
				name = L["Width"],
				desc = L["BarWidth_Desc"],	
				type = "range",
				min = 20,
				max = 100,
				step = 1,
				set = function(info, val)
				    self.db.profile.width = val
				end,
				get = function(info, val)
				    return self.db.profile.width
				end,
			},
			height = {
				order = 320,
				name = L["Height"],
				desc = L["BarHeight_Desc"],
				type = "range",
				min = 10,
				max = 50,
				step = 1,
				set = function(info, val)
				    self.db.profile.height = val 
				end,
				get = function(info, val)
				    return self.db.profile.height
				end,					
			},
		    fonts = {
		        order = 400,
		        type = "header",
		        name = L["Font"],
		    },
			font_size = {
				order = 410,
				name = L["Font size"],
				desc = L["Font size for the bars."],
				type = "range",
				min = 8,
				max = 30,
				step = 1,
				set = function(info, val) 
					self.db.profile.font_size = val
					self:ResetFonts()
				end,
				get = function(info,val) return self.db.profile.font_size end,
			},
			font_face = {
				order = 420,
				type = "select",
				name = L["Font"],
				desc = L["Font to use."],
				values = LSM:HashTable("font"),
				dialogControl = 'LSM30_Font',
				get = function() return self.db.profile.font_face end,
				set = function(info, val) 
				    self.db.profile.font_face = val
					self:ResetFonts()
				end
			},
			font_outline = {
				name = L["Outline"],
				desc = L["FontOutlineDesc"],
				type = "toggle",
				order = 430,
				set = function(info, val)
				    self.db.profile.fontFlags.OUTLINE = val
				    self:ResetFonts()
				end,
                get = function(info)
                    return self.db.profile.fontFlags.OUTLINE
                end,
			},
			font_thickoutline = {
				name = L["Thick Outline"],
				desc = L["FontThickOutline_OptionDesc"],
				type = "toggle",
				order = 440,
				set = function(info, val)
				    self.db.profile.fontFlags.THICKOUTLINE = val
				    self:ResetFonts()
				end,
                get = function(info)
                    return self.db.profile.fontFlags.THICKOUTLINE
                end,
			},
			font_monochrome = {
				name = L["Monochrome"],
				desc = L["FontMonochromeDesc"],
				type = "toggle",
				order = 450,
				set = function(info, val)
				    self.db.profile.fontFlags.MONOCHROME = val
				    self:ResetFonts()
				end,
                get = function(info)
                    return self.db.profile.fontFlags.MONOCHROME
                end,
			},
            textures = {
                order = 500,
                type = "header",
                name = L["Textures"],
            },
			bartexture = {
				order = 510,
				name = L["Bar Texture"],
				desc = L["BarTexture_OptionDesc"],
				type = "select",
				values = LSM:HashTable("statusbar"),
				dialogControl = 'LSM30_Statusbar',
				get = function()
				    return self.db.profile.texture
				end,
				set = function(info, val)
				    self.db.profile.texture = val
					self:UpdateTextures()
				end,
			},
			bordertexture = {
				order = 520,
				name = L["Border"],
				desc = L["Border_OptionDesc"],
				type = "select",
				values = LSM:HashTable("border"),
				dialogControl = 'LSM30_Border',
				get = function()
				    return self.db.profile.border
				end,
				set = function(info, val)
				    self.db.profile.border = val
					self:SetMainFrameBorder()
				end,
			},
		},
	}
	return core
end

function CompactRunes:GetProcsOptions()
	local procs = {
	    order = 2,
		name = L["Procs"],
		type = "group",
		args = {
		    description = {
		        order = 1,
		        type = "description",
		        name = L["ProcsDesc"],
		    },
		    procsHdr = {
		        order = 100,
		        type = "header",
		        name = L["Procs"],
		    },
		},
	}

	for k, v in pairs(procsList) do
		procs.args[v:gsub(" ", "_")] = {
			name = SpellNames[v],
			desc = SpellNames[v],
			type = "toggle",
			order = 100 + k,
			set = function(info, val)
			    self.db.profile.enabled[v] = val
			end,
	        get = function(info)
				return self.db.profile.enabled[v]
			end,
		}
	end

	return procs		
end
		
function CompactRunes:OnInitialize()
    -- Load the settings
    self.db = LibStub("AceDB-3.0"):New("CompactRunesDB", defaults, "Default")

	-- Register for profile callbacks
	self.db.RegisterCallback(self, "OnProfileChanged", "Reset")
	self.db.RegisterCallback(self, "OnProfileCopied", "Reset")
	self.db.RegisterCallback(self, "OnProfileReset", "Reset")
	
	LSM:Register("statusbar", "Compact Runes", [[Interface\Addons\CompactRunes\CompactRunes]])

	icon:Register("CompactRunesLDB", Broker.obj, self.db.profile.minimap)
	LSM.RegisterCallback(CompactRunes, "LibSharedMedia_Registered")
end

function CompactRunes:Reset()
	self:ResetFonts()
	self:UpdateTextures()
	self:SetMainFrameBorder()
end

function CompactRunes:LibSharedMedia_Registered(event, mediatype, key)
	if strlen(self.db.profile.font_face) > 1 and mediatype == "font" then
		if self.db.profile.font_face == key then
			self:ResetFonts()
		end
	end
	if mediatype == "statusbar" then
	    self:UpdateTextures()
	end
end

function CompactRunes:ResetFonts()
	for k, v in pairs(self.bars) do
		if v then
			v:ResetFonts()
		end
	end

	frames.procInd.ResetFonts()
end

function CompactRunes:UpdateTextures()
	for k, v in pairs(self.bars) do
		if v then
			v:UpdateTexture()
		end
	end
end

function CompactRunes:ChatCommand(input)
    if not input or input:trim() == "" then
        self:ShowOptions()
	elseif input:trim() == "debug" then
		self.db.profile.debug = true
		self:Print("Debug enabled.")
	elseif input:trim() == "nodebug" then
		self.db.profile.debug = false
		self:Print("Debug disabled.")
	end
end

function CompactRunes:OnEnable()
    self:CheckTalents()
	self:Load()
	LoadSpellNames()

	if not self.optionsFrame then
		-- Register Options
	    local displayName = GetAddOnMetadata(ADDON_NAME, "Title")
		local options = self:GetOptions()
	    LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable(displayName, options)

	    self.optionsFrame = {}
	    local ACD = LibStub("AceConfigDialog-3.0")
		self.optionsFrame.Main = ACD:AddToBlizOptions(
		    displayName, displayName, nil, "core")
		self.optionsFrame.Procs = ACD:AddToBlizOptions(
		    displayName, L["Procs"], displayName, "procs")
		ACD:AddToBlizOptions(
		    displayName, options.args.profile.name, displayName, "profile")
	    -- Register the chat command
	    self:RegisterChatCommand("crunes", "ChatCommand")
	end
end

local WatchedEvents = {
	"RUNE_POWER_UPDATE",
	"RUNE_TYPE_UPDATE",
	"UNIT_POWER",
	"UNIT_MAXPOWER",
	"PLAYER_REGEN_DISABLED",
	"PLAYER_REGEN_ENABLED",
	"PLAYER_TARGET_CHANGED",
	"UNIT_AURA",
	"UNIT_SPELLCAST_SUCCEEDED",
	"PLAYER_TALENT_UPDATE",
	"PLAYER_SPECIALIZATION_CHANGED",
	"COMBAT_LOG_EVENT_UNFILTERED",
}
function CompactRunes:Load()
	if self.isDK then
		self:CreateDisplay()
		for i = 1, #WatchedEvents do
			self:RegisterEvent(WatchedEvents[i])
		end
	end
end

function CompactRunes:Unload()
end

function CompactRunes:OnDisable()
	for i = 1, #WatchedEvents do
		self:UnregisterEvent(WatchedEvents[i])
	end
end

function CompactRunes:CheckTalents()
    local class, className = UnitClass("player")
    if className then
        if (className == 'DEATH KNIGHT' or className == 'DEATHKNIGHT') then
            self.isDK = true

			for i = 1, 18 do
				local name, iconTexture, tier, column, selected, available = 
					GetTalentInfo(i, "player")
				if name and selected and available then
					if name == SpellNames["Blood Tap"] then
						self.regenMethod = "BT"
					elseif name == SpellNames["Runic Empowerment"] then
						self.regenMethod = "RE"
					elseif name == SpellNames["Runic Corruption"] then
						self.regenMethod = "RC"
					else
						self.regenMethod = "None"
					end		
				end
			end
        else
            self.isDK = false
			self.regenMethod = "None"
        end
    end
	if self.db.profile.debug then
		local fmt = "CheckTalents [isDK=%s,regen=%s]"
		self:Print(
			fmt:format(tostring(self.isDK),
				self.regenMethod))
	end
end

local function onUpdateRune(self, elapsed)
	self.showing.timer = self.showing.timer + elapsed
	self:SetValue(self.showing.timer)
	if CompactRunes.db.profile.runeRechargeNums then
		local remain = abs(ceil(self.showing.duration - self.showing.timer))
		if self.showing.start > 0 then
			self.value:SetText(tostring(remain))
		else
			self.value:SetText("")
		end
	end
end

local function onUpdateDisease(self, elapsed)
	if self.active then
		self.timer = self.timer - elapsed
		if self.timer < 0 then self.timer = 0 end
		self:Show()
		self:SetValue(self.timer)
		self.value:SetText(tostring(floor(self.timer)))
	else
		self.value:SetText("0")
		self:Hide()
	end
end

local function onUpdateDarkTrans(self, elapsed)
	if self.active then
		self.timer = self.timer - elapsed
		if self.timer < 0 then self.timer = 0 end
		self:Show()
		self:SetValue(self.timer)
		if self.dt then
			self.value:SetText(tostring(floor(self.timer)))
		end
	else
		self:Hide()
	end
end

local fontFlags = {}
function CompactRunes:GetFontSettings()
	wipe(fontFlags)
	for k, v in pairs(self.db.profile.fontFlags) do
		if v then tinsert(fontFlags, k) end
	end
	return self.db.profile.font_face, self.db.profile.font_size, tconcat(fontFlags, ",")
end

-- Define a generic class for the bars
Bar.__index = Bar

function Bar:Create(name, fontAdj, width, height, setPoint, setValuePoint)
    local object = setmetatable({}, Bar)
	object.name = name
	object.fontAdj = fontAdj
	
	local widthFunc = width
	if type(width) ~= "function" then
		widthFunc = function() return width end
	end
	object.GetWidth = widthFunc
	local heightFunc = width
	if type(height) ~= "function" then
		heightFunc = function() return height end
	end
	object.GetHeight = height

	object.SetPoint = setPoint
	object.SetValuePoint = setValuePoint
	object:Initialize()
	-- Add the bar to the addon's table of bars
	CompactRunes.bars[name] = object
	return object
end

function Bar:Initialize()
    local bar = CreateFrame("StatusBar", "CompactRunes_"  .. self.name, frames.main)
	self.bar = bar
	self.db = CompactRunes.db.profile.bars[self.name]
	self.altcolor = false
	self.isRuneBar = self.name:find("RuneBar%d")
	bar:SetScale(1)
    bar:SetOrientation("HORIZONTAL")
    bar:SetWidth(self.GetWidth())
    bar:SetHeight(self.GetHeight())
	local bt = LSM:Fetch("statusbar", CompactRunes.db.profile.texture)
    bar:SetStatusBarTexture(bt)
    bar:GetStatusBarTexture():SetHorizTile(false)
    bar:GetStatusBarTexture():SetVertTile(false)
	if not self.isRuneBar then
    	local bc = self.db.color
    	bar:SetStatusBarColor(bc.r, bc.g, bc.b, bc.a)
	end
    bar.bg = bar:CreateTexture(nil, "BACKGROUND")
    bar.bg:SetTexture(bt)
    bar.bg:SetAllPoints(true)
	if not self.isRuneBar then
		local bgc = self.db.bgcolor
    	bar.bg:SetVertexColor(bgc.r, bgc.g, bgc.b, bgc.a)
	end

	local ff, fh, fflags = CompactRunes:GetFontSettings()
	local font = LSM:Fetch("font", ff)
    bar.value = bar:CreateFontString(nil, "OVERLAY")
    bar.value:SetPoint("CENTER")
    bar.value:SetFont(font, fh + self.fontAdj, fflags)
    bar.value:SetJustifyH("CENTER")
    --bar.value:SetShadowOffset(1, -1)
   	local tc = self.db.textcolor
   	bar.value:SetTextColor(tc.r, tc.g, tc.b, tc.a)
	if self.isRuneBar and not CompactRunes.db.profile.runeRechargeNums then
		bar.value:Hide()
	end

	if self.SetPoint then self:SetPoint() end
	if self.SetValuePoint then
		self:SetValuePoint()
	else
	    bar.value:SetPoint("CENTER")
	end

	if self.name == "FrostFeverBar" or 
		self.name == "BloodPlagueBar" or self.name == "DarkTransformationBar" then
		bar.active = false
		bar.timer = 0
		bar:Hide()
	end
end

local numberFmt = "%.0f"
function Bar:SetValue(value)
	self.bar:SetValue(value)
	self.bar.value:SetText(numberFmt:format(value))
end

function Bar:Reset()
	--self.bar:SetPoint("CENTER", UIParent, "CENTER", self.db.x, self.db.y)
	self:UpdateTexture()
	self:UpdateVisibility()
	self:UpdateGraphics()
end

function Bar:ResetFonts()
	local ff, fh, fontFlags = CompactRunes:GetFontSettings()
	local font = LSM:Fetch("font", ff)
	self.bar.value:SetFont(font, fh + self.fontAdj, fontFlags)						
	self.bar.value:SetText(self.bar.value:GetText())
end

function Bar:UpdateVisibility()
	self:UpdateTexture()
end

function Bar:UpdateTexture()
    bt = LSM:Fetch("statusbar", CompactRunes.db.profile.texture)
	self.bar:SetStatusBarTexture(bt)
	self.bar.bg:SetTexture(bt)
    self.bar:GetStatusBarTexture():SetHorizTile(false)
    self.bar:GetStatusBarTexture():SetVertTile(false)
	--self:UpdateGraphics()
end

function Bar:UpdateGraphics()
    local bc, bgc, tc

	if not self.isRuneBar then
		if self.altcolor then
		    bc = self.db.alt_color or self.db.color
		    bgc = self.db.alt_bgcolor or self.db.bgcolor
		    tc = self.db.alt_textcolor or self.db.textcolor
		else
		    bc = self.db.color
		    bgc = self.db.bgcolor
		    tc = self.db.textcolor
		end

	    self.bar:SetStatusBarColor(bc.r, bc.g, bc.b, bc.a)
	    self.bar.bg:SetVertexColor(bgc.r, bgc.g, bgc.b, bgc.a)
	    self.bar.value:SetTextColor(tc.r, tc.g, tc.b, tc.a)
	end
end

local runes = {
	[1] = {
		partner = 2,
		timer = 0,
		duration = 10,
		start = 0,
		type = 1,
		ready = true,
	},
	[2] = {
		partner = 1,
		timer = 0,
		duration = 10,
		start = 0,
		type = 1,
		ready = true,
	},
	[3] = {
		partner = 4,
		timer = 0,
		duration = 10,
		start = 0,
		type = 1,
		ready = true,
	},
	[4] = {
		partner = 3,
		timer = 0,
		duration = 10,
		start = 0,
		type = 1,
		ready = true,
	},
	[5] = {
		partner = 6,
		timer = 0,
		duration = 10,
		start = 0,
		type = 1,
		ready = true,
	},
	[6] = {
		partner = 5,
		timer = 0,
		duration = 10,
		start = 0,
		type = 1,
		ready = true,
	},
}
local runeBars = {}
local orderings = {
	["BUF"] = { 1, 2, 3, 4, 5, 6 },
	["BFU"] = { 1, 2, 5, 6, 3, 4 },
	["UFB"] = { 3, 4, 5, 6, 1, 2 },
	["FBU"] = { 5, 6, 1, 2, 3, 4 },
}
function CompactRunes:SetRuneOrder(ordering)
	for i = 1, 6 do
		local barNum = orderings[ordering][i]
		local bar = runeBars[barNum].bar
		bar:ClearAllPoints()
	    bar:SetPoint("TOPLEFT", frames.main, "TOPLEFT", 
			self.db.profile.runes.inset + (self.db.profile.width + self.db.profile.spacing) * ((i - 1) % 2), 
			- self.db.profile.runes.inset - (self.db.profile.height + self.db.profile.spacing) * (ceil(i / 2) - 1))
	end
end

function CompactRunes:UpdateRuneType(runeIndex)
	local partnerIndex = runes[runeIndex].partner
	local bar = runeBars[runeIndex]
	local type = GetRuneType(runeIndex)
	runes[runeIndex].type = type

	local time = GetTime()
	local firstIndex = min(runeIndex, partnerIndex)
	local secondIndex = max(runeIndex, partnerIndex)
	local first = runeBars[firstIndex]
	local second = runeBars[secondIndex]
	local sooner, later = runes[firstIndex], runes[secondIndex]
	if time - sooner.start < time - later.start then
		sooner, later = runes[secondIndex], runes[firstIndex]
	end

	local rc, rcc
	-- Update first
	rc = RuneColors[RuneMapping[sooner.type]]
	rcc = RuneCooldownColors[RuneMapping[sooner.type]]
    first.bar:SetStatusBarColor(rc.r, rc.g, rc.b, 1)
	first.bar.bg:SetVertexColor(rcc.r, rcc.g, rcc.b, rcc.a)
	first.bar.showing = sooner
	-- Update second
	rc = RuneColors[RuneMapping[later.type]]
	rcc = RuneCooldownColors[RuneMapping[later.type]]
    second.bar:SetStatusBarColor(rc.r, rc.g, rc.b, 1)
	second.bar.bg:SetVertexColor(rcc.r, rcc.g, rcc.b, rcc.a)
	second.bar.showing = later
end

function CompactRunes:UpdateRuneCooldown(runeIndex, isEnergized)
	local partnerIndex = runes[runeIndex].partner
	local start, duration, ready = GetRuneCooldown(runeIndex)
	local type = GetRuneType(runeIndex)
	local rune = runes[runeIndex]
	local bar = runeBars[runeIndex]
	rune.ready = ready

	if rune.ready then
		local timeLeft = rune.timer - rune.duration
		if timeLeft > 0.5 then
			self.TotalStats.reProcs[type] = (self.TotalStats.reProcs[type] or 0) + 1
			self.LastFightStats.reProcs[type] = (self.LastFightStats.reProcs[type] or 0) + 1
			if self.db.profile.debug then
				self:Print("RE Proc - " .. (RuneMapping[type] or "Unknown"))
			end
		end
		--local refFmt = "Rune %d: %0.2f [%0.2f, %0.2f]"
		--self:Print(refFmt:format(runeIndex, timeLeft, 
		--	rune.duration, rune.timer))
	end

	rune.start = start
	if ready == false then
		rune.duration = duration
		rune.timer = GetTime() - start
	else
		rune.timer = 10
		rune.duration = 10
	end

	local time = GetTime()
	local firstIndex = min(runeIndex, partnerIndex)
	local secondIndex = max(runeIndex, partnerIndex)
	local first = runeBars[firstIndex]
	local second = runeBars[secondIndex]
	local sooner, later = runes[firstIndex], runes[secondIndex]
	if time - sooner.start < time - later.start then
		sooner, later = runes[secondIndex], runes[firstIndex]
	end

	-- Update first
	first.bar.showing = sooner
	first.bar:SetMinMaxValues(0, sooner.duration)
	first.bar:SetValue(sooner.timer)
	if sooner.ready == false then
		first.bar:SetScript("OnUpdate", onUpdateRune)
		first.bar:SetAlpha(self.db.profile.rechargeAlpha)
	else
		first.bar:SetScript("OnUpdate", nil)
		first.bar:SetAlpha(1.0)
		if CompactRunes.db.profile.runeRechargeNums then
			first.bar.value:SetText("")
		end
	end
	-- Update second
	second.bar.showing = later
	second.bar:SetMinMaxValues(0, later.duration)
	second.bar:SetValue(later.timer)
	if later.ready == false then
		second.bar:SetScript("OnUpdate", onUpdateRune)
		second.bar:SetAlpha(self.db.profile.rechargeAlpha)
	else
		second.bar:SetScript("OnUpdate", nil)
		second.bar:SetAlpha(1.0)
		if CompactRunes.db.profile.runeRechargeNums then
			second.bar.value:SetText("")
		end
	end
end

function CompactRunes:CreateProcIndicator()
    local bar = CreateFrame("Frame", "CompactRunes_Proc", frames.main)
	bar.db = self.db.profile
	frames.procInd = bar
    bar:SetPoint("BOTTOM", frames.dtBar.bar, "TOP", 0, self.db.profile.spacing * 2)
	bar:SetScale(1)
    bar:SetWidth(bar.db.height * 3)
    bar:SetHeight(bar.db.height * 3)
	bar.texture = bar:CreateTexture(nil, "ARTWORK")
    bar.texture:SetAllPoints(bar)
    bar.texture:SetVertexColor(1, 1, 1)
	local ff, fh, fflags = self:GetFontSettings()
	local font = LSM:Fetch("font", ff)
    bar.value = bar:CreateFontString(nil, "OVERLAY")
    bar.value:SetPoint("BOTTOMRIGHT", bar, "BOTTOMRIGHT", 0, -5)
    bar.value:SetFont(font, fh + 3, fflags)
    bar.value:SetJustifyH("CENTER")
	bar.value:SetJustifyV("CENTER")
    bar.value:SetShadowOffset(1, -1)
    local tc = bar.db.bars.ProcInd.textcolor
    bar.value:SetTextColor(tc.r, tc.g, tc.b, tc.a)
	bar.value:Hide()
	
	bar.ResetFonts = function()
		local ff, fh, fontFlags = self:GetFontSettings()
		local font = LSM:Fetch("font", ff)
		bar.value:SetFont(font, fh + 3, fontFlags)						
		bar.value:SetText(bar.value:GetText())
	end
end

function CompactRunes:SetMainFrameBorder()
	local frame = frames.main
	local border = LSM:Fetch("border", self.db.profile.border)
	local backdrop = {
        bgFile = "Interface\\Buttons\\WHITE8X8", 
        edgeFile = border, tile = false, tileSize = 16, edgeSize = 16, 
        insets = { left = 4, right = 4, top = 4, bottom = 4 }
    }
	local bdbc = self.db.profile.frame.bdcolor
	local bdc = self.db.profile.frame.color
    frame:SetBackdrop(backdrop)
    frame:SetBackdropBorderColor(bdbc.r, bdbc.g, bdbc.b, bdbc.a)
    frame:SetBackdropColor(bdc.r, bdc.g, bdc.b, bdc.a)
end

function CompactRunes:CalculateSize()
	self.sizes = self.sizes or {}
	self.sizes.width = 0
	self.sizes.height = 0
	self.sizes.hinset = 0
	self.sizes.vinset = 0
end

function CompactRunes:CreateDisplay()
	local frame = CreateFrame("Frame", "CompactRunes_Display", UIParent)
	frames.main = frame
	frame.db = self.db.profile.frame
	frame:SetPoint("CENTER", UIParent, "CENTER", frame.db.x, frame.db.y)
	frame:SetWidth((self.db.profile.width + self.db.profile.spacing) * 2 + 9)
	frame:SetHeight((self.db.profile.height + self.db.profile.spacing) * 3 + 9)
	self:SetMainFrameBorder()
	frame.lock = self.db.profile.locked
    frame:SetMovable()
    frame:RegisterForDrag("LeftButton")
    frame:SetScript("OnDragStart",
        function(self, button)
			if not self.lock then
            	self:StartMoving()
			end
        end)
    frame:SetScript("OnDragStop",
        function(self)
            self:StopMovingOrSizing()
			local scale = self:GetEffectiveScale() / UIParent:GetEffectiveScale()
			local x, y = self:GetCenter()
			x, y = x * scale, y * scale
			x = x - GetScreenWidth()/2
			y = y - GetScreenHeight()/2
			x = x / self:GetScale()
			y = y / self:GetScale()
			self.db.x, self.db.y = x, y
			self:SetUserPlaced(false);
        end)
    frame:EnableMouse(not self.lock)

	frames.powerBar = Bar:Create("RunicPowerBar", 2, 
		function() return self.db.profile.width * 2 + self.db.profile.spacing end,
		function() return self.db.profile.height end,
		function(self)
			self.bar:SetPoint("TOPLEFT", frames.main, "BOTTOMLEFT", 
				CompactRunes.db.profile.runes.inset, -2)
		end
	)
	self:UpdateRunicPowerMax()
	self:UpdateRunicPower()

	frames.ffBar = Bar:Create("FrostFeverBar", 2, 
		function() return self.db.profile.width end,
		function() return self.db.profile.height end,
		function(self)
			self.bar:SetPoint("BOTTOMLEFT", frames.main, "TOPLEFT", 
				CompactRunes.db.profile.runes.inset, 0)
		end,
		function(self) self.bar.value:SetPoint("CENTER", self.bar, "TOP", 0, 0) end
	)

	frames.bpBar = Bar:Create("BloodPlagueBar", 2, 
		function() return self.db.profile.width end,
		function() return self.db.profile.height end,
		function(self)
			self.bar:SetPoint("TOPLEFT", frames.ffBar.bar, "TOPRIGHT", 
				CompactRunes.db.profile.spacing, 0)
		end,
		function(self) self.bar.value:SetPoint("CENTER", self.bar, "TOP", 0, 0) end
	)

	frames.dtBar = Bar:Create("DarkTransformationBar", 2, 
		function() return self.db.profile.width end,
		function() return self.db.profile.height end,
		function(self)
			self.bar:SetPoint("BOTTOM", frames.main, "TOP", 
				0, CompactRunes.db.profile.spacing * 2 + 
					CompactRunes.db.profile.height + 2)
		end,
		function(self) self.bar.value:SetPoint("CENTER", self.bar, "TOP", 0, 0) end
	)

	self:CreateProcIndicator()
	
	-- Create the runes and then update their status
	for i = 1, 6 do
		runeBars[i] = Bar:Create("RuneBar"..i, 0, 
			function() return self.db.profile.width end,
			function() return self.db.profile.height end,
			nil,
			nil
		)
		runeBars[i].bar.showing = runes[i]
	end
	for i = 1, 6 do
		self:UpdateRuneType(i)
		self:UpdateRuneCooldown(i)
	end	
	self:SetRuneOrder(self.db.profile.runeOrder)
	frames.main:SetAlpha(self.db.profile.ooc_alpha)
	frames.main:Show()
end

function CompactRunes:RUNE_POWER_UPDATE(event, runeIndex, isEnergized)
	self:UpdateRuneCooldown(runeIndex, isEnergized)
end

function CompactRunes:RUNE_TYPE_UPDATE(event, runeIndex)
	self:UpdateRuneType(runeIndex)
end

function CompactRunes:UNIT_POWER(event, unit, type)
	if unit == "player" then -- and type == "RUNIC_POWER" then
		self:UpdateRunicPower()
	end
end

function CompactRunes:UpdateRunicPower()
	frames.powerBar:SetValue(UnitPower("player", 6))
end

function CompactRunes:UpdateRunicPowerMax()
	frames.powerBar.bar:SetMinMaxValues(0, UnitPowerMax("player", 6))
end

function CompactRunes:UNIT_MAXPOWER(event)
	self:UpdateRunicPowerMax()
end

function CompactRunes:PLAYER_REGEN_DISABLED()
	-- Player has entered combat.
	frames.main:SetAlpha(1)

	self.LastFightStats:Reset()
end

function CompactRunes:PLAYER_REGEN_ENABLED()
	-- Player has left combat.
	frames.main:SetAlpha(self.db.profile.ooc_alpha)
end

function CompactRunes:PLAYER_TARGET_CHANGED(unit)
	self:UNIT_AURA("UNIT_AURA", "target")
end

local procPriority = {
	SpellNames["Freezing Fog"],
	SpellNames["Killing Machine"],
	SpellNames["Sudden Doom"],
	SpellNames["Blood Charge"],
}
local petBuffs = {
	SpellNames["Dark Transformation"],
	SpellNames["Shadow Infusion"],
}
local diseases
function CompactRunes:UNIT_AURA(event, unit)
	if unit == "target" then
		diseases = diseases or {
			[SpellNames["Frost Fever"]] = frames.ffBar,
			[SpellNames["Blood Plague"]] = frames.bpBar,
		}

		local name, rank, icon, count, dispelType, duration, expires, 
			caster, isStealable, shouldConsolidate, spellID, canApplyAura, 
			isBossDebuff, value1, value2, value3 

		for spell, bar in pairs(diseases) do
			name, rank, icon, count, dispelType, duration, expires, 
			caster, isStealable, shouldConsolidate, spellID, canApplyAura, 
			isBossDebuff, value1, value2, value3 
			= UnitDebuff("target", spell, nil, "PLAYER")

			if name and bar.db.enabled then
				bar.bar.active = true 
				bar.bar.timer = expires - GetTime()
				bar.bar:SetMinMaxValues(0, duration)
				bar.bar:Show()
				bar.bar:SetScript("OnUpdate", onUpdateDisease)
			else
				bar.bar.active = false 
				bar.bar.timer = 0
				bar.bar:SetScript("OnUpdate", nil)
				bar.bar:Hide()
			end
		end
	elseif unit == "player" then
		if self.db.profile.procs then
			local procInd = frames.procInd
			for i, spell in pairs(procPriority) do
				name, rank, icon, count, dispelType, duration, expires, 
				caster, isStealable, shouldConsolidate, spellID, canApplyAura, 
				isBossDebuff, value1, value2, value3 
				= UnitBuff("player", spell)
				if name then break end
			end

			if name and self.db.profile.enabled[name] ~= false then
				procInd.texture:SetTexture(icon)
				procInd:Show()
				if count and count > 1 then
					procInd.value:SetText(tostring(count))
					procInd.value:Show()
				else
					procInd.value:Hide()
				end
			else
				procInd:Hide()
			end
		end
	elseif unit == "pet" then
		for i, spell in pairs(petBuffs) do
			name, rank, icon, count, dispelType, duration, expires, 
			caster, isStealable, shouldConsolidate, spellID, canApplyAura, 
			isBossDebuff, value1, value2, value3 
			= UnitBuff("pet", spell, nil, "PLAYER")
			if name then break end
		end

		local dtBar = frames.dtBar
		if name then
			dtBar.bar.dt = (name == SpellNames["Dark Transformation"])
			dtBar.altcolor = dtBar.bar.dt
			dtBar.bar.timer = expires - GetTime()
			local dtdb = self.db.profile.bars.DarkTransformationBar
			local bc, tc
			if dtBar.altcolor then
				bc = dtdb.alt_color
				tc = dtdb.alt_textcolor
			else
				bc = dtdb.color
				tc = dtdb.textcolor
				dtBar.bar.value:SetText(tostring(count))
			end
			dtBar.bar:SetMinMaxValues(0, duration)
			dtBar.bar:SetStatusBarColor(bc.r, bc.g, bc.b, bc.a)
			dtBar.bar.value:SetTextColor(tc.r, tc.g, tc.b, tc.a)
			if self.db.profile.bars.DarkTransformationBar.enabled then
				dtBar.bar.active = true 
				dtBar.bar:Show()
				dtBar.bar:SetScript("OnUpdate", onUpdateDarkTrans)
			end
		else
			dtBar.bar.active = true 
			dtBar.bar.timer = 0
			dtBar.bar:SetScript("OnUpdate", nil)
			dtBar.bar:Hide()
		end
	end
end

function CompactRunes:UNIT_SPELLCAST_SUCCEEDED(event, unit, spellName)
    if unit == "player" then
		if spellName == SpellNames["Death Coil"] or
			spellName == SpellNames["Rune Strike"] or
			spellName == SpellNames["Frost Strike"] then

				local regenMethod = self.regenMethod
				self.TotalStats.regenUses[regenMethod] = (self.TotalStats.regenUses[regenMethod] or 0) + 1
				self.LastFightStats.regenUses[regenMethod] = (self.LastFightStats.regenUses[regenMethod] or 0) + 1
		end
	end
end

function CompactRunes:PLAYER_SPECIALIZATION_CHANGED(event)
	self:CheckTalents()
end

function CompactRunes:PLAYER_TALENT_UPDATE(event)
	self:CheckTalents()
end

function CompactRunes:COMBAT_LOG_EVENT_UNFILTERED(...)
    local event, timestamp, eventtype, hideCaster, 
        srcGUID, srcName, srcFlags, srcRaidFlags, 
        destGUID, destName, destFlags, destRaidFlags, 
        param9, param10, param11, param12, param13, param14, 
        param15, param16, param17, param18, param19, param20

    event, timestamp, eventtype, hideCaster, 
    srcGUID, srcName, srcFlags, srcRaidFlags,
    destGUID, destName, destFlags, destRaidFlags,
    param9, param10, param11, param12, param13, param14, 
    param15, param16, param17, param18, param19, param20 = ...

	if eventtype == "SPELL_AURA_APPLIED" and destName == self.playerName then
		if param9 == SpellIds["Runic Corruption"] then
			self.LastFightStats.rcProcs = self.LastFightStats.rcProcs + 1
			self.TotalStats.rcProcs = self.TotalStats.rcProcs + 1
		end
	end
end
